/******************************************************************************
 *
 * Filename: cycdefs.c
 *
 * Author:   Robert Thornburrow
 * Project:  Cycle - The CYbot Control LanguagE
 * Created:  5th June 2003
 *
 * Purpose:  Parse tree structure manipulation functions definitions.
 *
 * (c) Copyright R.H.Thornburrow, 2003
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 *****************************************************************************/

#include "config.h"

#if defined(WIN32) && CRTMEMCHECK
#include <crtdbg.h>
extern _CrtMemState crts1, crts2, crts3;
#endif /* WIN32 && CRTMEMCHECK */

#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#include "cycdefs.h"
#include "errors.h"
#include "hashtable.h"

extern const char* infile;
extern const char* exepath;
extern int errors;
extern int warnings;
extern int _lineNo;


/**
 * Map of class names to Class structures.
 */
Hashtable *classTable = NULL;

/**
 * Map of instance names to Instance structures.
 */
Hashtable *instanceTable = NULL;

/**
 * Map of procedure names to Procedure structures.
 */
Hashtable *procedureTable = NULL;


/******************************************************************************
 * Lists
 */

ClassList* newClassList( Class *el, ClassList *l )
{
	ClassList *ret = (ClassList*) malloc( sizeof( ClassList ) );
	ret->head = el;
	ret->tail = l;
	return ret;
}


MemberList* newMemberList( Member *el, MemberList *l )
{
	MemberList *ret = (MemberList*) malloc( sizeof( MemberList ) );
	ret->head = el;
	ret->tail = l;
	return ret;
}


void freeMemberList( MemberList *mbrlist )
{
	MemberList *temp;

	for( ; mbrlist; mbrlist = temp )
	{
		if( mbrlist->head )
			freeMember( mbrlist->head );

		temp = mbrlist->tail;
		free( mbrlist );
	}
}


InstanceList* newInstanceList( Instance *el, InstanceList *l )
{
	InstanceList *ret = (InstanceList*) malloc( sizeof( InstanceList ) );
	ret->head = el;
	ret->tail = l;
	return ret;
}


ProcedureList *newProcedureList( Procedure *el, ProcedureList *l )
{
	ProcedureList *ret = (ProcedureList*) malloc( sizeof( ProcedureList ) );
	ret->head = el;
	ret->tail = l;
	return ret;
}


IdentList *newIdentList( Ident *el, IdentList *l )
{
	IdentList *ret = (IdentList*) malloc( sizeof( IdentList ) );
	ret->head = el;
	ret->tail = l;
	return ret;
}


void freeIdentList( IdentList *idlist )
{
	IdentList *temp;

	for( ; idlist; idlist = temp )
	{
		if( idlist->head )
			freeIdent( idlist->head );

		temp = idlist->tail;
		free( idlist );
	}
}


StatementList *newStatementList( Statement *el, StatementList *l )
{
	StatementList *ret = (StatementList*) malloc( sizeof( StatementList ) );
	ret->head = el;
	ret->tail = l;
	return ret;
}


void freeStatementList( StatementList *stmtlist )
{
	StatementList *temp;

	for( ; stmtlist; stmtlist = temp )
	{
		if( stmtlist->head )
			freeStatement( stmtlist->head );

		temp = stmtlist->tail;
		free( stmtlist );
	}
}


ExpressionList *newExpressionList( Expression *el, ExpressionList *l )
{
	ExpressionList *ret = ( ExpressionList *) malloc( sizeof( ExpressionList ) );
	ret->head = el;
	ret->tail = l;
	return ret;
}


void freeExpressionList( ExpressionList *exprlist )
{
	ExpressionList *temp;

	for( ; exprlist; exprlist = temp )
	{
		if( exprlist->head )
			freeExpression( exprlist->head );

		temp = exprlist->tail;
		free( exprlist );
	}
}


CaseBlockList *newCaseBlockList( CaseBlock *head, CaseBlockList *tail )
{
	CaseBlockList *ret = ( CaseBlockList *) malloc( sizeof( CaseBlockList ) );
	ret->head = head;
	ret->tail = tail;
	return ret;
}


PICList *newPICList( PIC *el, PICList *l )
{
	PICList *ret = ( PICList*) malloc( sizeof( PICList ) );
	ret->head = el;
	ret->tail = l;
	return ret;
}


LineList *newLineList( Line *el, LineList *l )
{
	LineList *ret = ( LineList*) malloc( sizeof( LineList ) );
	ret->head = el;
	ret->tail = l;
	return ret;
}


/******************************************************************************
 * Classes, Identifiers, Procedures and Programs
 */

Class* newClass( Ident* name, IdentList *pArgs, struct MemberList *pMembers )
{
	MemberList *mbrlist;
	Class *ret;

	/* Make sure class name is not "main" as this is a reserved name */
	if( strcmp( "main", name->name ) == 0 )
	{
		semantic_error( "\"main\" is not a valid class name.", name->lineNo );
		freeIdent( name );
		if( pArgs )  freeIdentList( pArgs );
		if( pMembers )  freeMemberList( pMembers );
		return NULL;
	}

	/* Does a class with this name already exist? */
	if( classTable ? Hashtable_get( classTable, name->name ) : NULL )
	{
		semantic_error1( "re-definition of class '%s'.", name->lineNo, name->name );
		freeIdent( name );
		if( pArgs )  freeIdentList( pArgs );
		if( pMembers )  freeMemberList( pMembers );
		return NULL;
	}

	/* Does an instance or procedure with this name already exist? */
	if( (instanceTable ? Hashtable_get( instanceTable, name->name ) : NULL)
		|| (procedureTable ? Hashtable_get( procedureTable, name->name ) : NULL) )
	{
		semantic_error1( "re-definition of '%s'.", name->lineNo, name->name );
		freeIdent( name );
		if( pArgs )  freeIdentList( pArgs );
		if( pMembers )  freeMemberList( pMembers );
		return NULL;
	}

	ret = (Class*) malloc( sizeof( Class ) );
	ret->name        = name;
	ret->pArgs		 = pArgs;
	ret->pMembers    = pMembers;
	ret->memberTable = NULL;

	/* Add the new class to the table of classes */
	if( !classTable )
		classTable = Hashtable_new( 11 );

	Hashtable_put( classTable, name->name, ret );

	/* Build a hashtable of all the members of the class */
	for( mbrlist = pMembers; mbrlist; mbrlist = mbrlist->tail )
	{
		if( mbrlist->head )
		{
			if( !ret->memberTable )
				ret->memberTable = Hashtable_new( 11 );

			/* Does a member with this name already exist? */
			if( Hashtable_get( ret->memberTable , mbrlist->head->name->name ) )
			{
				semantic_error1( "redefinition of member %s.", mbrlist->head->name->lineNo, mbrlist->head->name->name );
				return NULL;
			}

			/* Add the member to the members hashtable for this class */
			Hashtable_put( ret->memberTable, mbrlist->head->name->name, mbrlist->head );
		}
	}

	return ret;
}


void freeClass( Class *clazz )
{
	if( clazz->memberTable )
	{
		Hashtable_delete( clazz->memberTable, 0, 0, 0, NULL );
		clazz->memberTable = NULL;
	}
	if( clazz->name )
	{
		freeIdent( clazz->name );
		clazz->name = NULL;
	}
	if( clazz->pArgs )
	{
		freeIdentList( clazz->pArgs );
		clazz->pArgs = NULL;
	}
	if( clazz->pMembers )
	{
		freeMemberList( clazz->pMembers );
		clazz->pMembers = NULL;
	}

	free( clazz );
}


LineNumber* newLineNumber( const char* filename, int line )
{
	LineNumber* ret = (LineNumber*) malloc( sizeof( LineNumber ) );
	ret->filename = filename;
	ret->line     = line;
	return ret;
}


LineNumber* copyLineNumber( LineNumber *lineNo )
{
	LineNumber* ret = (LineNumber*) malloc( sizeof( LineNumber ) );
	ret->filename = lineNo->filename;
	ret->line     = lineNo->line;
	return ret;
}


void freeLineNumber( LineNumber *lineNo )
{
	/* filename is a shared string, so don't free it */
	lineNo->filename = NULL;
	free( lineNo );
}


Ident* newIdent( char* name, IdentType type )
{
	Ident* ret = (Ident*) malloc( sizeof( Ident ) );
	ret->name   = name;
	ret->type   = type;
	ret->lineNo = newLineNumber( infile, _lineNo );
	return ret;
}


Ident* copyIdent( Ident *ident )
{
	Ident* ret = (Ident*) malloc( sizeof( Ident ) );
	ret->name   = strdup( ident->name );
	ret->type   = ident->type;
	ret->lineNo = copyLineNumber( ident->lineNo );
	return ret;
}


void freeIdent( Ident *ident )
{
	free( (char*) ident->name );
	ident->name = NULL;

	if( ident->lineNo )
	{
		freeLineNumber( ident->lineNo );
		ident->lineNo = NULL;
	}

	free( ident );
}


Procedure* newProcedure( Ident* name, IdentList* pArgs, struct StatementList *pStmts )
{
	Procedure* ret;

	/* Make sure procedure with this name has not already been defined */
	if( procedureTable ? Hashtable_get( procedureTable, name->name ) : NULL )
	{
		semantic_error1( "re-definition of procedure '%s'.", name->lineNo, name->name );
		freeIdent( name );
		if( pArgs )  freeIdentList( pArgs );
		if( pStmts )  freeStatementList( pStmts );
		return NULL;
	}

	/* Make sure procedure has a different name from a previously defined instance or class */
	if( (instanceTable ? Hashtable_get( instanceTable, name->name ) : NULL)
		|| (classTable ? Hashtable_get( classTable, name->name ) : NULL) )
	{
		semantic_error1( "re-definition of '%s'.", name->lineNo, name->name );
		freeIdent( name );
		if( pArgs )  freeIdentList( pArgs );
		if( pStmts )  freeStatementList( pStmts );
		return NULL;
	}

	ret = (Procedure*) malloc( sizeof( Procedure ) );
	ret->name   = name;
	ret->pArgs  = pArgs;
	ret->pStmts = pStmts;

	/* Add the new class to the table of classes */
	if( !procedureTable )
		procedureTable = Hashtable_new( 11 );

	Hashtable_put( procedureTable, name->name, ret );

	return ret;
}


void freeProcedure( Procedure *proc )
{
	if( proc->name )
	{
		freeIdent( proc->name );
		proc->name = NULL;
	}
	if( proc->pArgs )
	{
		freeIdentList( proc->pArgs );
		proc->pArgs = NULL;
	}
	if( proc->pStmts )
	{
		freeStatementList( proc->pStmts );
		proc->pStmts = NULL;
	}

	free( proc );
}


Program* newProgram( ProcedureList *pp )
{
	Program* ret = (Program*) malloc( sizeof( Program ) );
	ret->pProcs = pp;
	ret->sourcefile = NULL;
	ret->outputfile = NULL;
	return ret;
}


/******************************************************************************
 * Members
 */

MemberDef* newMemberDef( MemberDefType kind, struct ExpressionList *pArgs )
{
	MemberDef* ret = (MemberDef*) malloc( sizeof( MemberDef ) );
	ret->kind  = kind;
	ret->pArgs = pArgs;
	return ret;
}


void freeMemberDef( MemberDef *mbrdef )
{
	if( mbrdef->pArgs )
	{
		freeExpressionList( mbrdef->pArgs );
		mbrdef->pArgs = NULL;
	}

	free( mbrdef );
}


Range* newRange( Expression *from, Expression *to )
{
	Range* ret = (Range*) malloc( sizeof( Range ) );
	ret->from  = from;
	ret->to    = to;
	return ret;
}


void freeRange( Range *range )
{
	if( range->from )
	{
		freeExpression( range->from );
		range->from = NULL;
	}
	if( range->to )
	{
		freeExpression( range->to );
		range->to = NULL;
	}

	free( range );
}


Member* newConstMember( Ident *name, struct Expression *value )
{
	Member* ret = (Member*) malloc( sizeof( Member ) );
	ret->kind   = const_mbr;
	ret->name   = name;
	ret->parent = NULL;
	ret->data.const_part.value = value;
	return ret;
}


Member* newMember( MemberType kind, Ident *name, IdentList *pArgs, MemberDef *def )
{
	Member* ret = (Member*) malloc( sizeof( Member ) );
	ret->kind   = kind;
	ret->name   = name;
	ret->parent = NULL;
	ret->data.normal_part.pArgs = pArgs;
	ret->data.normal_part.definition = def;
	return ret;
}


Member* newInputMember( Ident *name, IdentList *pArgs, MemberDef *def, Range *range )
{
	Member* ret = (Member*) malloc( sizeof( Member ) );
	ret->kind   = input_mbr;
	ret->name   = name;
	ret->parent = NULL;
	ret->data.input_part.pArgs      = pArgs;
	ret->data.input_part.definition = def;
	ret->data.input_part.range      = range;
	return ret;
}


void freeMember( Member *member )
{
	if( member->name )
	{
		freeIdent( member->name );
		member->name = NULL;
	}

	member->parent = NULL;

	switch( member->kind )
	{
	case const_mbr:
		if( member->data.const_part.value )
		{
			freeExpression( member->data.const_part.value );
			member->data.const_part.value = NULL;
		}
		break;
	case input_mbr:
		if( member->data.input_part.definition )
		{
			freeMemberDef( member->data.input_part.definition );
			member->data.input_part.definition = NULL;
		}
		if( member->data.input_part.pArgs )
		{
			freeIdentList( member->data.input_part.pArgs );
			member->data.input_part.pArgs = NULL;
		}
		if( member->data.input_part.range )
		{
			freeRange( member->data.input_part.range );
			member->data.input_part.range = NULL;
		}
		break;
	case process_mbr:
	case output_mbr:
		if( member->data.normal_part.definition )
		{
			freeMemberDef( member->data.normal_part.definition );
			member->data.normal_part.definition = NULL;
		}
		if( member->data.normal_part.pArgs )
		{
			freeIdentList( member->data.normal_part.pArgs );
			member->data.normal_part.pArgs = NULL;
		}
		break;
	}

	free( member );
}


/******************************************************************************
 * Instances
 */

Instance *newInstance( Ident *type, Ident *name, ExpressionList *pArgs )
{
	Instance* ret;

	/* Lookup the class object for the instance type */
	Class *clazz = classTable ? Hashtable_get( classTable, type->name ) : NULL;

	if( !clazz )
	{
		semantic_error1( "Undeclared class %s in instance declaration.", type->lineNo, type->name );
		freeIdent( type );
		freeIdent( name );
		if( pArgs )  freeExpressionList( pArgs );
		return NULL;
	}

	/* Make sure instance name is not "main" as this is a reserved name */
	if( strcmp( "main", name->name ) == 0 )
	{
		semantic_error( "\"main\" is not a valid instance name.", name->lineNo );
		freeIdent( type );
		freeIdent( name );
		if( pArgs )  freeExpressionList( pArgs );
		return NULL;
	}

	/* Does an instance with this name already exist? */
	if( instanceTable ? Hashtable_get( instanceTable, name->name ) : NULL )
	{
		semantic_error1( "re-definition of instance '%s'.", name->lineNo, name->name );
		freeIdent( type );
		freeIdent( name );
		if( pArgs )  freeExpressionList( pArgs );
		return NULL;
	}

	/* Make sure instance has a different name from a previously defined class or procedure */
	if( (classTable ? Hashtable_get( classTable, name->name ) : NULL)
		|| (procedureTable ? Hashtable_get( procedureTable, name->name ) : NULL) )
	{
		semantic_error1( "re-definition of '%s'.", name->lineNo, name->name );
		freeIdent( type );
		freeIdent( name );
		if( pArgs )
			freeExpressionList( pArgs );
		return NULL;
	}

	ret = (Instance*) malloc( sizeof( Instance ) );
	ret->name  = name;
	ret->type  = type;
	ret->pArgs = pArgs;
	ret->clazz = clazz;

	if( !instanceTable )
		instanceTable = Hashtable_new( 11 );

	Hashtable_put( instanceTable, name->name, ret );

	return ret;
}


void freeInstance( Instance *inst )
{
	/* Just a reference to a real Class structure */
	inst->clazz = NULL;

	if( inst->name )
	{
		freeIdent( inst->name );
		inst->name = NULL;
	}
	if( inst->pArgs )
	{
		freeExpressionList( inst->pArgs );
		inst->pArgs = NULL;
	}
	if( inst->type )
	{
		freeIdent( inst->type );
		inst->type = NULL;
	}

	free( inst );
}


/******************************************************************************
 * Statements
 */


Statement* newIfStatement( struct Expression *c,
  struct StatementList *pt, struct StatementList *pe )
{
	Statement* ret = (Statement*) malloc( sizeof( Statement ) );
	ret->kind   = if_stmt;
	ret->lineNo = newLineNumber( infile, _lineNo );
	ret->data.if_part.condition = c;
	ret->data.if_part.pThen = pt;
	ret->data.if_part.pElse = pe;
	return ret;
}


Statement* newWhileStatement( struct Expression *c, struct StatementList *pb )
{
	Statement* ret = (Statement*) malloc( sizeof( Statement ) );
	ret->kind   = while_stmt;
	ret->lineNo = newLineNumber( infile, _lineNo );
	ret->data.while_and_do_part.condition = c;
	ret->data.while_and_do_part.pBody = pb;
	return ret;
}


Statement* newDoStatement( struct Expression *c, struct StatementList *pb )
{
	Statement* ret = (Statement*) malloc( sizeof( Statement ) );
	ret->kind   = do_stmt;
	ret->lineNo = newLineNumber( infile, _lineNo );
	ret->data.while_and_do_part.condition = c;
	ret->data.while_and_do_part.pBody = pb;
	return ret;
}


Statement* newForStatement( Statement *init, struct Expression *cond, Statement *incr, StatementList *pbody )
{
	Statement* ret = (Statement*) malloc( sizeof( Statement ) );
	ret->kind   = for_stmt;
	ret->lineNo = newLineNumber( infile, _lineNo );
	ret->data.for_part.initial   = init;
	ret->data.for_part.condition = cond;
	ret->data.for_part.increment = incr;
	ret->data.for_part.pBody     = pbody;
	return ret;
}


Statement* newSwitchStatement( struct Expression *expr, CaseBlockList *pCaseBlocks )
{
	Statement* ret = (Statement*) malloc( sizeof( Statement ) );
	ret->kind   = switch_stmt;
	ret->lineNo = newLineNumber( infile, _lineNo );
	ret->data.switch_part.expr		  = expr;
	ret->data.switch_part.pCaseBlocks = pCaseBlocks;
	return ret;
}


CaseBlock* newCaseBlock( struct ExpressionList *pCases, StatementList *pStmts )
{
	CaseBlock* ret = (CaseBlock*) malloc( sizeof( CaseBlock ) );
	ret->pCases = pCases;
	ret->pStmts = pStmts;
	return ret;
}


void freeCaseBlock( CaseBlock *caseblock )
{
	if( caseblock->pCases )
	{
		freeExpressionList( caseblock->pCases );
		caseblock->pCases = NULL;
	}
	if( caseblock->pStmts )
	{
		freeStatementList( caseblock->pStmts );
		caseblock->pStmts = NULL;
	}
	free( caseblock );
}


Statement* newAssignmentStatement( Ident *name, AssignmentType kind, struct Expression *value )
{
	Statement* ret = (Statement*) malloc( sizeof( Statement ) );
	ret->kind   = assignment_stmt;
	ret->lineNo = newLineNumber( infile, _lineNo );
	ret->data.assignment_part.kind  = kind;
	ret->data.assignment_part.name  = name;
	ret->data.assignment_part.value = value;
	return ret;
}


Statement* newCallStatement( Ident* clazz, Ident *name, struct ExpressionList *pArgs )
{
	Statement* ret;

	/* Is this a procedure call or a member call? */
	if( clazz )
	{
		Member *member;

		/* Check clazz is an instance of a class */
		Instance *inst = instanceTable ? Hashtable_get( instanceTable, clazz->name ) : NULL;

		if( !inst )
		{
			semantic_error1( "'%s' is not an instance of a class.", clazz->lineNo, clazz->name );
			freeIdent( clazz );
			freeIdent( name );
			if( pArgs )  freeExpressionList( pArgs );
			return NULL;
		}

		/* Check clazz is linked back to a Class object */
		if( !inst->clazz )
		{
			semantic_error1( "Class '%s' not linked.", clazz->lineNo, clazz->name );
			freeIdent( clazz );
			freeIdent( name );
			if( pArgs )  freeExpressionList( pArgs );
			return NULL;
		}

		/* Check that name is a member of the clazz's class */
		member = inst->clazz->memberTable ? Hashtable_get( inst->clazz->memberTable, name->name ) : NULL;

		if( !member )
		{
			semantic_error2( "Class '%s' does not contain a definition for '%s'.", clazz->lineNo, clazz->name, name->name );
			freeIdent( clazz );
			freeIdent( name );
			if( pArgs )  freeExpressionList( pArgs );
			return NULL;
		}

		/* Check that member is a output or process member (not a constant) */
		if( (member->kind != output_mbr) && (member->kind != process_mbr) )
		{
			semantic_error2( "'%s.%s' is not a output or process member.", clazz->lineNo, clazz->name, name->name );
			freeIdent( clazz );
			freeIdent( name );
			if( pArgs )  freeExpressionList( pArgs );
			return NULL;
		}

		ret = (Statement*) malloc( sizeof( Statement ) );
		ret->kind   = call_stmt;
		ret->lineNo = newLineNumber( infile, _lineNo );
		ret->data.call_part.clazz = clazz;
		ret->data.call_part.name  = name;
		ret->data.call_part.pArgs = pArgs;
		ret->data.call_part.definition.member = member;
	}
	else	/* procedure call */
	{
		ret = (Statement*) malloc( sizeof( Statement ) );
		ret->kind   = call_stmt;
		ret->lineNo = newLineNumber( infile, _lineNo );
		ret->data.call_part.clazz = NULL;
		ret->data.call_part.name  = name;
		ret->data.call_part.pArgs = pArgs;
	}
	return ret;
}


Statement* newGotoStatement( struct Expression *e )
{
	Statement* ret = (Statement*) malloc( sizeof( Statement ) );
	ret->kind   = goto_stmt;
	ret->lineNo = NULL;
	ret->data.goto_and_label_part.expr = e;
	return ret;
}


Statement* newLabelStatement( struct Expression *e )
{
	Statement* ret = (Statement*) malloc( sizeof( Statement ) );
	ret->kind   = label_stmt;
	ret->lineNo = NULL;
	ret->data.goto_and_label_part.expr = e;
	return ret;
}


Statement* newSimpleStatement( int kind )
{
	Statement* ret = (Statement*) malloc( sizeof( Statement ) );
	ret->kind   = kind;
	ret->lineNo = newLineNumber( infile, _lineNo );
	return ret;
}


void freeStatement( Statement *stmt )
{
	if( stmt->lineNo )
	{
		freeLineNumber( stmt->lineNo );
		stmt->lineNo = NULL;
	}

	switch( stmt->kind )
	{
	case if_stmt:
		if( stmt->data.if_part.condition )
		{
			freeExpression( stmt->data.if_part.condition );
			stmt->data.if_part.condition = NULL;
		}
		if( stmt->data.if_part.pThen )
		{
			freeStatementList( stmt->data.if_part.pThen );
			stmt->data.if_part.pThen = NULL;
		}
		if( stmt->data.if_part.pElse )
		{
			freeStatementList( stmt->data.if_part.pElse );
			stmt->data.if_part.pElse = NULL;
		}
		break;
	case while_stmt:
	case do_stmt:
		if( stmt->data.while_and_do_part.condition )
		{
			freeExpression( stmt->data.while_and_do_part.condition );
			stmt->data.while_and_do_part.condition = NULL;
		}
		if( stmt->data.while_and_do_part.pBody )
		{
			freeStatementList( stmt->data.while_and_do_part.pBody );
			stmt->data.while_and_do_part.pBody = NULL;
		}
		break;
	case for_stmt:
		if( stmt->data.for_part.initial )
		{
			freeStatement( stmt->data.for_part.initial );
			stmt->data.for_part.initial = NULL;
		}
		if( stmt->data.for_part.condition )
		{
			freeExpression( stmt->data.for_part.condition );
			stmt->data.for_part.condition = NULL;
		}
		if( stmt->data.for_part.increment )
		{
			freeStatement( stmt->data.for_part.increment );
			stmt->data.for_part.increment = NULL;
		}
		if( stmt->data.for_part.pBody )
		{
			freeStatementList( stmt->data.for_part.pBody );
			stmt->data.for_part.pBody = NULL;
		}
		break;
	case switch_stmt:
		if( stmt->data.switch_part.expr )
		{
			freeExpression( stmt->data.switch_part.expr );
			stmt->data.switch_part.expr = NULL;
		}
		if( stmt->data.switch_part.pCaseBlocks )
		{
			CaseBlockList *caseblocklist, *temp;
			for( caseblocklist = stmt->data.switch_part.pCaseBlocks; caseblocklist; caseblocklist = temp )
			{
				if( caseblocklist->head )
					freeCaseBlock( caseblocklist->head );

				temp = caseblocklist->tail;
				free( caseblocklist );
			}
			stmt->data.switch_part.pCaseBlocks = NULL;
		}
		break;
	case assignment_stmt:
		if( stmt->data.assignment_part.name )
		{
			freeIdent( stmt->data.assignment_part.name );
			stmt->data.assignment_part.name = NULL;
		}
		if( stmt->data.assignment_part.value )
		{
			freeExpression( stmt->data.assignment_part.value );
			stmt->data.assignment_part.value = NULL;
		}
		break;
	case call_stmt:
		if( stmt->data.call_part.clazz )
		{
			freeIdent( stmt->data.call_part.clazz );
			stmt->data.call_part.clazz = NULL;
		}
		stmt->data.call_part.definition.proc = NULL;
		if( stmt->data.call_part.name )
		{
			freeIdent( stmt->data.call_part.name );
			stmt->data.call_part.name = NULL;
		}
		if( stmt->data.call_part.pArgs )
		{
			freeExpressionList( stmt->data.call_part.pArgs );
			stmt->data.call_part.pArgs = NULL;
		}
		break;
	case goto_stmt:
	case label_stmt:
		if( stmt->data.goto_and_label_part.expr )
		{
			freeExpression( stmt->data.goto_and_label_part.expr );
			stmt->data.goto_and_label_part.expr = NULL;
		}
		break;
	case return_stmt:
	case break_stmt:
	case start_stmt:
		break;
	default:
		fprintf( stderr, "WARNING: unknown statement kind (%d) during cleanup.", stmt->kind );
		break;
	}

	free( stmt );
}


/******************************************************************************
 * Expressions
 */

Expression* newVariableExpr( Ident *v )
{
    Expression* ret = (Expression*) malloc( sizeof( Expression ) );
    ret->kind               = variable_expr;
	ret->lineNo				= newLineNumber( infile, _lineNo );
	ret->type               = v->type;
    ret->data.variable_part = v;
	return ret;
}


Expression* newIntegerExpr( int k, int n )
{
	Expression* ret = (Expression*) malloc( sizeof( Expression ) );
	ret->kind              = k;
	ret->lineNo				= newLineNumber( infile, _lineNo );
	ret->type              = integer_type;
	ret->data.integer_part = n;
	return ret;
}


Expression* newStringExpr( const char *s )
{
    Expression* ret = (Expression*) malloc( sizeof( Expression ) );
    ret->kind              = string_expr;
	ret->lineNo				= newLineNumber( infile, _lineNo );
	ret->type              = string_type;
    ret->data.string_part  = s;
	return ret;
}


Expression* newBinaryOpExpression( ExprKind k, Expression* l, Expression* r )
{
	Expression* ret;

	/* Check to see if there was a problem looking up a constant for example! */
	if( !l || !r )
	{
		if( l )  freeExpression( l );
		if( r )  freeExpression( r );
		return NULL;
	}

	ret = (Expression*) malloc( sizeof( Expression ) );
	ret->kind   = k;
	ret->lineNo	= newLineNumber( infile, _lineNo );
	ret->data.binary_op_part.left = l;
	ret->data.binary_op_part.right = r;

	/* Expression type checking code:
     *
     * addition, subtraction, multiplication, division = type of params
     * and, or = boolean or integer depending in context!
     * equality, inequality, lessthan, greaterthan, lessequal, greaterequal = boolean
     *
     * parameter types work on simple promotion and checks against conversion
     * of strings.
	 */

	switch( k )
	{
    case equality_expr: case inequality_expr:
    case lessthan_expr: case greaterthan_expr:
	case lessequal_expr: case greaterequal_expr:
		ret->type = boolean_type;
		return ret;
	}

	if( l->type == r->type )
	{
		ret->type = l->type;
		return ret;
	}

	switch( l->type )
	{
	case integer_type:
		switch( r->type )
		{
		case string_type:  printf( "ERROR: Can't convert string to integer\n" ); break;
        case boolean_type: printf( "WARNING: Conversion of integer to boolean\n" ); ret->type = boolean_type; return ret;
		}
		break;
	case string_type:
		switch( r->type )
		{
		case integer_type: printf( "ERROR: Can't convert string to integer\n" ); break;
        case boolean_type: printf( "ERROR: Can't convert string to boolean\n" ); break;
		}
		break;
    case boolean_type:
		switch( r->type )
		{
		case integer_type: printf( "WARNING: Conversion of integer to boolean\n" ); ret->type = boolean_type; return ret;
		case string_type:  printf( "ERROR: Can't convert string to boolean\n" ); break;
		}
		break;
	}

	ret->type = untyped_id;
    return ret;
}


Expression* newCallExpression( Ident* clazz, Ident* name, ExpressionList* pArgs )
{
	Instance *inst;
	Member *member;
	Expression* ret;

	/* Must be an member */
	if( !clazz )
	{
		semantic_error1( "'%s' is not a member.", name->lineNo, name->name );
		freeIdent( name );
		if( pArgs )  freeExpressionList( pArgs );
		return NULL;
	}

	/* Check clazz is an instance of a class */
	inst = instanceTable ? Hashtable_get( instanceTable, clazz->name ) : NULL;

	if( !inst )
	{
		semantic_error1( "'%s' is not an instance of a class.", clazz->lineNo, clazz->name );
		freeIdent( clazz );
		freeIdent( name );
		if( pArgs )  freeExpressionList( pArgs );
		return NULL;
	}

	/* Check clazz is linked back to a Class object */
	if( !inst->clazz )
	{
		semantic_error1( "Class '%s' not linked.", clazz->lineNo, clazz->name );
		freeIdent( clazz );
		freeIdent( name );
		if( pArgs )  freeExpressionList( pArgs );
		return NULL;
	}

	/* Check that name is a member of the clazz's class */
	member = inst->clazz->memberTable ? Hashtable_get( inst->clazz->memberTable, name->name ) : NULL;

	if( !member )
	{
		semantic_error2( "Class '%s' does not contain a definition for '%s'.", clazz->lineNo, clazz->name, name->name );
		freeIdent( clazz );
		freeIdent( name );
		if( pArgs )  freeExpressionList( pArgs );
		return NULL;
	}

	/* Check that member is an input member */
	if( member->kind != input_mbr )
	{
		semantic_error2( "'%s.%s' is not an input member.", clazz->lineNo, clazz->name, name->name );
		freeIdent( clazz );
		freeIdent( name );
		if( pArgs )  freeExpressionList( pArgs );
		return NULL;
	}

	ret = (Expression*) malloc( sizeof( Expression ) );
	ret->kind   = call_expr;
	ret->lineNo	= newLineNumber( infile, _lineNo );
	ret->type   = name->type;
	ret->data.call_part.clazz = clazz;
	ret->data.call_part.name  = name;
	ret->data.call_part.pArgs = pArgs;
	ret->data.call_part.definition = member;
	return ret;
}


Expression* newConstantExpression( Ident* clazz, Ident* name )
{
	Expression* ret;
	Class *theClass;
	Member *member;

	/* Is this a procedure call or a member call? */
	if( !clazz )
	{
		semantic_error1( "'%s' must be qualified by a class name.", name->lineNo, name->name );
		freeIdent( name );
		return NULL;
	}


	/* Check clazz is an instance of a class */
	theClass = classTable ? Hashtable_get( classTable, clazz->name ) : NULL;

	if( !theClass )
	{
		semantic_error1( "'%s' is not a class.", clazz->lineNo, clazz->name );
		freeIdent( clazz );
		freeIdent( name );
		return NULL;
	}

	/* Check that name is a member of the clazz's class */
	member = theClass->memberTable ? Hashtable_get( theClass->memberTable, name->name ) : NULL;

	if( !member )
	{
		semantic_error2( "Class '%s' does not contain a definition for '%s'.", clazz->lineNo, clazz->name, name->name );
		freeIdent( clazz );
		freeIdent( name );
		return NULL;
	}

	if( member->kind != const_mbr )
	{
		semantic_error2( "'%s.%s' is not a constant member.", clazz->lineNo, clazz->name, name->name );
		freeIdent( clazz );
		freeIdent( name );
		return NULL;
	}

	ret = (Expression*) malloc( sizeof( Expression ) );
	ret->kind   = constant_expr;
	ret->lineNo	= newLineNumber( infile, _lineNo );
	ret->type   = member->data.const_part.value->type;
	ret->data.constant_part.clazz = clazz;
	ret->data.constant_part.name  = name;
	ret->data.constant_part.value = member->data.const_part.value;
	return ret;
}


Expression* newDefaultExpr( const char *s )
{
    Expression* ret = (Expression*) malloc( sizeof( Expression ) );
    ret->kind   = default_expr;
	ret->lineNo	= newLineNumber( infile, _lineNo );
	ret->type   = untyped_id;
	ret->data.default_part.pCases = NULL;
	return ret;
}


void freeExpression( Expression *expr )
{
	if( expr->lineNo )
	{
		freeLineNumber( expr->lineNo );
		expr->lineNo = NULL;
	}

	switch( expr->kind )
	{
	case variable_expr:
		if( expr->data.variable_part )
		{
			freeIdent( expr->data.variable_part );
			expr->data.variable_part = NULL;
		}
		break;
	case string_expr:
		if( expr->data.string_part )
		{
			free( (char*) expr->data.string_part );
			expr->data.string_part = NULL;
		}
		break;
	case boolean_expr:
	case integer_expr:
		/* Nothing to free */
		break;
	case addition_expr:
	case subtraction_expr:
	case multiplication_expr:
	case division_expr:
	case equality_expr:
	case inequality_expr:
	case lessthan_expr:
	case greaterthan_expr:
	case lessequal_expr:
	case greaterequal_expr:
		if( expr->data.binary_op_part.left )
		{
			freeExpression( expr->data.binary_op_part.left );
			expr->data.binary_op_part.left = NULL;
		}
		if( expr->data.binary_op_part.right )
		{
			freeExpression( expr->data.binary_op_part.right );
			expr->data.binary_op_part.right = NULL;
		}
		break;
	case call_expr:
		if( expr->data.call_part.clazz )
		{
			freeIdent( expr->data.call_part.clazz );
			expr->data.call_part.clazz = NULL;
		}
		expr->data.call_part.definition = NULL;
		if( expr->data.call_part.name )
		{
			freeIdent( expr->data.call_part.name );
			expr->data.call_part.name = NULL;
		}
		if( expr->data.call_part.pArgs )
		{
			freeExpressionList( expr->data.call_part.pArgs );
			expr->data.call_part.pArgs = NULL;
		}
		break;
	case constant_expr:
		if( expr->data.constant_part.clazz )
		{
			freeIdent( expr->data.constant_part.clazz );
			expr->data.constant_part.clazz = NULL;
		}
		if( expr->data.constant_part.name )
		{
			freeIdent( expr->data.constant_part.name );
			expr->data.constant_part.name = NULL;
		}
		/* value is owned by the constant member - this is just a link to it! */
		expr->data.constant_part.value = NULL;
		break;
	case default_expr:
		if( expr->data.default_part.pCases )
		{
			freeExpressionList( expr->data.default_part.pCases );
			expr->data.default_part.pCases = NULL;
		}
		break;
	default:
		fprintf( stderr, "WARNING: unknown expression kind (%d) during cleanup.", expr->kind );
		break;
	}

	free( expr );
}


/******************************************************************************
 * PIC (Program declaration - Procedure, Instance or Class)
 */

PIC *newProcedurePIC( Procedure *proc )
{
	PIC *ret = (PIC*) malloc( sizeof( PIC ) );
	ret->kind = proc_pic;
	ret->data.proc = proc;
	return ret;
}


PIC *newInstancePIC( Instance *instance )
{
	PIC *ret = (PIC*) malloc( sizeof( PIC ) );
	ret->kind = instance_pic;
	ret->data.instance = instance;
	return ret;
}


PIC *newClassPIC( Class *clazz )
{
	PIC *ret = (PIC*) malloc( sizeof( PIC ) );
	ret->kind = class_pic;
	ret->data.clazz = clazz;
	return ret;
}


void freePIC( PIC *pic )
{
	/* Data owned by class, procedure and instance lists */
	free( pic );
}


/******************************************************************************
 * Lines
 */

Line* newLine( int fid, int fport, int tid, int tport )
{
	Line* ret = (Line*) malloc( sizeof( Line ) );
	ret->fromid   = fid;
	ret->fromport = fport;
	ret->toid     = tid;
	ret->toport   = tport;
	return ret;
}

void freeLine( Line *line )
{
	free( line );
}


/******************************************************************************
 * Utility
 */

const char* typeToString( ExprType t )
{
	switch( t )
	{
	case integer_type: return "integer";
	case string_type:  return "string";
	case boolean_type: return "boolean";
	case untyped_type:
	default:
		return "<untyped>";
	}
}


/* Wrapper to fix type cast in Hashtable_delete */
void freeProcedureWrapper( void *proc )
{
	freeProcedure( (Procedure*) proc );
}


/* Wrapper to fix type cast in Hashtable_delete */
void freeInstanceWrapper( void *inst )
{
	freeInstance( (Instance*) inst );
}


/* Wrapper to fix type cast in Hashtable_delete */
void freeClassWrapper( void *clazz )
{
	freeClass( (Class*) clazz );
}


void cleanup( Program *prog )
{
	ProcedureList *proclist, *tempproc;
	InstanceList *instlist, *tempinst;
	ClassList *classlist, *tempclass;
	PICList *piclist, *temppic;
	LineList *linelist, *templine;

	/* Have created any procedures?  If so, free them all, regardless of whether
	   they have been added to the program */
	if( procedureTable )
	{
		verbose_msg( infile, "Freeing procedures..." );
		Hashtable_delete( procedureTable, 0, 0, 0, &freeProcedureWrapper );
		procedureTable = NULL;
	}

	/* Free the linked list of procedures - the procedure structures have already gone */
	for( proclist = prog->pProcs; proclist; proclist = tempproc )
	{
		/* Free the linked list element */
		tempproc = proclist->tail;
		free( proclist );
	}

	prog->pProcs = NULL;

	/* Have created any instances?  If so, free them all, regardless of whether
	   they have been added to the program */
	if( instanceTable )
	{
		verbose_msg( infile, "Freeing instances..." );
		Hashtable_delete( instanceTable, 0, 0, 0, &freeInstanceWrapper );
		instanceTable = NULL;
	}

	/* Free the linked list of instances - the instance structures have already gone */
	for( instlist = prog->pInstances; instlist; instlist = tempinst )
	{
		/* Free the linked list element */
		tempinst = instlist->tail;
		free( instlist );
	}

	prog->pInstances = NULL;

	/* Have created any classes?  If so, free them all, regardless of whether
	   they have been added to the program */
	if( classTable )
	{
		verbose_msg( infile, "Freeing classes..." );
		Hashtable_delete( classTable, 0, 0, 0, &freeClassWrapper );
		classTable = NULL;
	}

	/* Free the linked list of classes - the class structures have already gone */
	for( classlist = prog->pClasses; classlist; classlist = tempclass )
	{
		/* Free the linked list element */
		tempclass = classlist->tail;
		free( classlist );
	}

	prog->pClasses = NULL;

	for( piclist = prog->pPICs; piclist; piclist = temppic )
	{
		if( piclist->head )
			freePIC( piclist->head );

		/* Free the linked list element */
		temppic = piclist->tail;
		free( piclist );
	}

	prog->pPICs = NULL;

	for( linelist = prog->pLines; linelist; linelist = templine )
	{
		if( linelist->head )
			freeLine( linelist->head );

		/* Free the linked list element */
		templine = linelist->tail;
		free( linelist );
	}

	prog->pPICs = NULL;

	free( (char*) prog->sourcefile );
	prog->sourcefile = NULL;

	free( (char*) prog->outputfile );
	prog->outputfile = NULL;

	free( (char*) infile );
	infile = NULL;

	free( (char*) exepath );
	exepath = NULL;

#if defined(WIN32) && CRTMEMCHECK
	printf( "Checking memory using CrtMemDifference()...\n" );
	_CrtMemCheckpoint( &crts2 );
	if ( _CrtMemDifference( &crts3, &crts1, &crts2 ) )
		_CrtMemDumpStatistics( &crts3 );
	_CrtCheckMemory();
#endif /* WIN32 && CRTMEMCHECK */
}
